@@ -1,11 +1,14 @@
-
 package URI::Fast;
 # ABSTRACT: A fast(er) URI parser
-$URI::Fast::VERSION = '0.03';
-use common::sense;
-use URI::Split qw(uri_split);
-use URI::Encode::XS qw(uri_decode uri_encode);
+$URI::Fast::VERSION = '0.04';
+
 use Carp;
+use Inline 'C';
+use URI::Encode::XS qw(uri_encode uri_decode);
+require Exporter;
+
+use parent 'Exporter';
+our @EXPORT_OK = qw(uri uri_split);
 
 use overload
   '""' => sub{
@@ -14,8 +17,6 @@ use overload
       . ($_[0]->{frag}  ? ('#' . $_[0]->{frag})  : '');
   };
 
-use parent 'Exporter';
-our @EXPORT_OK = qw(uri);
 
 # Regexes used to validate characters present in string when attributes are
 # updated.
@@ -26,7 +27,7 @@ our %LEGAL = (
 
 sub uri ($) {
   my $self = bless {}, __PACKAGE__;
-  @{$self}{qw(scheme auth path query frag)} = uri_split $_[0];
+  @{$self}{qw(scheme auth path query frag)} = uri_split($_[0]);
   $self->{scheme} //= 'file';
   $self;
 }
@@ -63,116 +64,71 @@ sub path {
   }
 }
 
-# For bits of the auth string, build a lazy accessor that calls _auth, which
-# parses the auth string.
+# For bits of the auth string, build a lazy accessor that parses the auth
+# string, as well as a setter which regenerates it.
 foreach my $attr (qw(usr pwd host port)) {
   *{__PACKAGE__ . "::$attr"} = sub {
-    $_[0]->{auth} && $_[0]->{_auth} || $_[0]->_auth;
+    # Parse auth section if necessary
+    unless ($_[0]->{auth} && $_[0]->{_auth}) {
+      $_[0]->{_auth} = {};
+      @{ $_[0]->{_auth} }{qw(usr pwd host port)} = auth_split($_[0]->{auth});
+      $_[0]->{_auth}{usr} = uri_decode($_[0]->{_auth}{usr}) if $_[0]->{_auth}{usr};
+      $_[0]->{_auth}{pwd} = uri_decode($_[0]->{_auth}{pwd}) if $_[0]->{_auth}{pwd};
+    }
 
     # Set new value, then regenerate authorization section string
     if (@_ == 2) {
       !exists($LEGAL{$attr}) || $_[1] =~ $LEGAL{$attr} || croak "illegal chars in $attr";
       $_[0]->{_auth}{$attr} = $_[1];
-      $_[0]->_reauth;
+
+      # Update auth string
+      $_[0]->{auth} = auth_join(
+        ($_[0]->{_auth}{usr} ? uri_encode($_[0]->{_auth}{usr}) : undef),
+        ($_[0]->{_auth}{pwd} ? uri_encode($_[0]->{_auth}{pwd}) : undef),
+        $_[0]->{_auth}{host},
+        $_[0]->{_auth}{port},
+      );
     }
 
     $_[0]->{_auth}{$attr};
   };
 }
 
-# Parses auth section
-sub _auth {
-  my ($self) = @_;
-  $self->{_auth} = {};                                     # Set a flag to prevent reparsing
-
-  if (local $_ = $self->auth) {
-    my ($cred, $loc) = split /@/;                          # usr:pwd@host:port
-
-    if ($loc) {
-      @{$self->{_auth}}{qw(usr pwd)}   = split ':', $cred; # Both credentials and location are present
-      @{$self->{_auth}}{qw(host port)} = split ':', $loc;
-    }
-    else {
-      @{$self->{_auth}}{qw(host port)} = split ':', $cred; # Only location is present
-    }
-  }
-}
-
-# Regenerates auth section
-sub _reauth {
-  my $self = shift;
-  $self->{auth} = '';
-
-  if ($self->{_auth}{usr}) {                                           # Add the credentials block (usr:pwd)
-    $self->{auth} = $self->{_auth}{usr};                               # Add user
-    $self->{auth} .= ':' . $self->{_auth}{pwd} if $self->{_auth}{pwd}; # Add :pwd if pwd present
-    $self->{auth} .= '@';
-  }
-
-  if ($self->{_auth}{host}) {
-    $self->{auth} .= $self->{_auth}{host};                             # Add host if present (may not be for, e.g. file://)
-
-    if ($self->{_auth}{port}) {
-      $self->{auth} .= ':' . $self->{_auth}{port};                     # Port only valid if host is present
-    }
-  }
-}
-
 sub param {
   my ($self, $key, $val) = @_;
   $self->{query} || $val || return;
 
-  if (!$self->{_query}) {
-    $self->{_query} = {};                                       # Somewhere to set our things
+  $key = uri_encode($key);
 
-    if (local $_ = $self->{query}) {                            # Faster access via a dynamic variable
-      tr/\+/ /;                                                 # Seriously, dfarrell?
-      local @_ = split /[&=]/;                                  # Tokenize
+  if ($val) {
+    # Wipe out any current values for $key
+    $self->{query} =~ s/&?\b$key=[^&#]+//g;
 
-      while (my $k = uri_decode(shift // '')) {                 # Decode the next self->{_query}eter
-        if ($val && $k eq $key) {                               # We'll be setting this shortly, so ignore
-          shift;
-          next;
-        }
-
-        if (exists $self->{_query}{$k}) {                       # Multiple self->{_query}eters exist with the same key
-          $self->{_query}{$k} = [$self->{_query}{$k}]           # Reinitialize as array ref to store multiple values
-            unless ref $self->{_query}{$k};
-
-          push @{$self->{_query}{$k}}, uri_decode(shift // ''); # Add to the array
-        }
-        else {                                                  # First or only use of key
-          $self->{_query}{$k} = uri_decode(shift // '');
-        }
-      }
-    }
-  };
-
-  if ($val) {                                                   # Modifying this value
-    $self->{_query}{$key} = $val;                               # Update the pre-parsed hash ref
-
-    $key = uri_encode($key);                                    # Encode the key
-    $self->{query} =~ s/\b$key=[^&]+&?//;                       # Remove from the query string
-
-    if (ref $val) {                                             # If $val is an array, each element gets its own place in the query
-      foreach (@$val) {
-        $self->{query} .= '&' if $self->{query};
-        $self->{query} .= $key . '=' . uri_encode($_);
-      }
-    }
-    else {                                                      # Otherwise, just add the encoded pair
+    # Encode and attach values for param to query string
+    foreach (ref $val ? @$val : ($val)) {
       $self->{query} .= '&' if $self->{query};
-      $self->{query} .= $key . '=' . uri_encode($val);
+      $self->{query} .= $key . '=' . uri_encode($_);
     }
   }
 
-  $self->{_query}{$key} if defined $key;
+  # Collect and decode values for $key
+  my @vals = $self->{query} =~ /\b$key=([^&#]+)/g;
+
+  if (@vals) {
+    if (@vals > 1) {
+      return map{ tr/+/ /; uri_decode($_) } @vals;
+    }
+    elsif ($vals[0]) {
+      $vals[0] =~ tr/+/ /;
+      return uri_decode($vals[0]);
+    }
+  }
+
+  return;
 }
 
 1;
 
-__END__
-
 =pod
 
 =encoding UTF-8
@@ -183,7 +139,7 @@ URI::Fast - A fast(er) URI parser
 
 =head1 VERSION
 
-version 0.03
+version 0.04
 
 =head1 SYNOPSIS
 
@@ -221,8 +177,7 @@ access and update portions of the URI.
 
 =head2 uri
 
-Accepts a string, minimally parses it (using L<URI::Split>), and returns a
-L<URI::Fast> object.
+Accepts a string, minimally parses it, and returns a L<URI::Fast> object.
 
 When initially created, only the scheme, authorization section, path, query
 string, and fragment are split out. Thesea are broken down in a lazy fashion
@@ -301,6 +256,20 @@ The fragment section of the URI, excluding the leading C<#>.
 
 See L<URI::Fast::Benchmarks>.
 
+=head1 SEE ALSO
+
+=over
+
+=item L<URI>
+
+The de facto standard.
+
+=item L<Panda::URI>
+
+Written in C++ and purportedly very fast, but appears to only support Linux.
+
+=back
+
 =head1 AUTHOR
 
 Jeff Ober <sysread@fastmail.fm>
@@ -313,3 +282,185 @@ This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
 =cut
+
+__DATA__
+__C__
+
+#include <string.h>
+
+void uri_split(SV* uri) {
+  STRLEN  len;
+  char*   src = SvPV(uri, len);
+  size_t  idx = 0;
+  size_t  brk = 0;
+
+  Inline_Stack_Vars;
+  Inline_Stack_Reset;
+
+  // Scheme
+  brk = strcspn(&src[idx], ":/@?#");
+  if (brk > 0 && strncmp(&src[idx + brk], "://", 3) == 0) {
+    Inline_Stack_Push(newSVpv(&src[idx], brk));
+    idx += brk + 3;
+
+    // Authority
+    brk = strcspn(&src[idx], "/?#");
+    if (brk > 0) {
+      Inline_Stack_Push(newSVpv(&src[idx], brk));
+      idx += brk;
+    } else {
+      Inline_Stack_Push(newSVpv("",0));
+    }
+  }
+  else {
+    Inline_Stack_Push(&PL_sv_undef);
+    Inline_Stack_Push(&PL_sv_undef);
+  }
+
+  // Path
+  brk = strcspn(&src[idx], "?#");
+  if (brk > 0) {
+    Inline_Stack_Push(newSVpv(&src[idx], brk));
+    idx += brk;
+  } else {
+    Inline_Stack_Push(newSVpv("",0));
+  }
+
+  // Query
+  if (src[idx] == '?') {
+    ++idx; // skip past ?
+    brk = strcspn(&src[idx], "#");
+    if (brk > 0) {
+      Inline_Stack_Push(newSVpv(&src[idx], brk));
+      idx += brk;
+    } else {
+      Inline_Stack_Push(&PL_sv_undef);
+    }
+  } else {
+    Inline_Stack_Push(&PL_sv_undef);
+  }
+
+  // Fragment
+  if (src[idx] == '#') {
+    ++idx; // skip past #
+    brk = len - idx;
+    if (brk > 0) {
+      Inline_Stack_Push(newSVpv(&src[idx], brk));
+    } else {
+      Inline_Stack_Push(&PL_sv_undef);
+    }
+  } else {
+    Inline_Stack_Push(&PL_sv_undef);
+  }
+
+  Inline_Stack_Done;
+}
+
+void auth_join(SV* usr, SV* pwd, SV* host, SV* port) {
+  STRLEN  len;
+  char*   tmp;
+  char    out[1024];
+  int     idx = 0;
+
+  Inline_Stack_Vars;
+  Inline_Stack_Reset;
+
+  if (SvTRUE(usr)) {
+    tmp = SvPV(usr, len);
+    strncpy(&out[idx], tmp, len);
+    idx += len;
+
+    if (SvTRUE(pwd)) {
+      out[idx++] = ':';
+
+      tmp = SvPV(pwd, len);
+      strncpy(&out[idx], tmp, len);
+      idx += len;
+    }
+
+    out[idx++] = '@';
+  }
+
+  if (SvTRUE(host)) {
+    tmp = SvPV(host, len);
+    strncpy(&out[idx], tmp, len);
+    idx += len;
+
+    if (SvTRUE(port)) {
+      out[idx++] = ':';
+
+      tmp = SvPV(port, len);
+      strncpy(&out[idx], tmp, len);
+      idx += len;
+    }
+  }
+
+  if (idx > 0) {
+    Inline_Stack_Push(newSVpv(out, idx));
+  } else {
+    Inline_Stack_Push(newSVpv("", 0));
+  }
+
+  Inline_Stack_Done;
+}
+
+void auth_split(SV* auth) {
+  STRLEN  len;
+  char*   src  = SvPV(auth, len);
+  size_t  idx  = 0;
+  size_t  brk1 = 0;
+  size_t  brk2 = 0;
+
+  SV* usr;
+  SV* pwd;
+
+  Inline_Stack_Vars;
+  Inline_Stack_Reset;
+
+  if (len == 0) {
+    Inline_Stack_Push(&PL_sv_undef);
+    Inline_Stack_Push(&PL_sv_undef);
+    Inline_Stack_Push(&PL_sv_undef);
+    Inline_Stack_Push(&PL_sv_undef);
+  }
+  else {
+    // Credentials
+    brk1 = strcspn(&src[idx], "@");
+
+    if (brk1 > 0 && brk1 != len) {
+      brk2 = strcspn(&src[idx], ":");
+
+      if (brk2 > 0 && brk2 < brk1) {
+        Inline_Stack_Push(newSVpv(&src[idx], brk2));
+        idx += brk2 + 1;
+
+        Inline_Stack_Push(newSVpv(&src[idx], brk1 - brk2 - 1));
+        idx += brk1 - brk2;
+      }
+      else {
+        Inline_Stack_Push(newSVpv(&src[idx], brk1));
+        Inline_Stack_Push(&PL_sv_undef);
+        idx += brk1 + 1;
+      }
+    }
+    else {
+      Inline_Stack_Push(&PL_sv_undef);
+      Inline_Stack_Push(&PL_sv_undef);
+    }
+
+    // Location
+    brk1 = strcspn(&src[idx], ":");
+
+    if (brk1 > 0 && brk1 != (len - idx)) {
+      Inline_Stack_Push(newSVpv(&src[idx], brk1));
+      idx += brk1 + 1;
+      Inline_Stack_Push(newSVpv(&src[idx], len - idx));
+    }
+    else {
+      Inline_Stack_Push(newSVpv(&src[idx], len - idx));
+      Inline_Stack_Push(&PL_sv_undef);
+    }
+  }
+
+  Inline_Stack_Done;
+}
